
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _MSC_VER 
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int   u32;
typedef unsigned __int64 u64;
#else
#include <stdint.h>
typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;
#endif

#ifdef _MSC_VER 
#define ROTL32 _rotl 
#define ROTL64 _rotl64
#else
#define ROTL32(x, c) (((x) << (c)) ^ ((x) >> (32-(c))))
#define ROTL64(x, c) (((x) << (c)) ^ ((x) >> (64-(c))))
#endif

#define SWAP32(v) \
  ((ROTL32(v,  8) & (0x00FF00FFU)) | \
   (ROTL32(v, 24) & (0xFF00FF00U)))

#define SWAP64(v) \
  ((ROTL64(v,  8) & (0x000000FF000000FFULL)) | \
   (ROTL64(v, 24) & (0x0000FF000000FF00ULL)) | \
   (ROTL64(v, 40) & (0x00FF000000FF0000ULL)) | \
   (ROTL64(v, 56) & (0xFF000000FF000000ULL)))
   
#ifdef LITTLE_ENDIAN
	#define GET_U64(x) (((u64*)(x))[0])
#elif BIG_ENDIAN
	#define GET_U64(x) ( SWAP64(((u64*)(x))[0]) )
#else
	#error Choose an endianness!
#endif

#define MIX(y1, y0, x1, x0, c) \
	(y0) = (x0) + (x1); \
	(y1) = ROTL64((x1), c) ^ (y0)
	
#define IMIX(x1, x0, y1, y0, c) \
	(x1) = ROTL64((y1) ^ (y0), 64-c); \
	(x0) = (y0) - (x1)
	
	
#define WORDS		(4) 
#define ROUNDS	(72)

static const u32 R[8][2] =
{
	{5, 56},
	{36,28},
	{13,46},
	{58,44},
	{26,20},
	{53,35},
	{11,42},
	{59,50}
};

static const u32 Pi[4] = {0, 3, 2, 1};

struct key_sched
{
	u64 subkey[ROUNDS/4+1][WORDS];
};

typedef struct key_sched ThreefishKey[1];

void ThreefishKS(ThreefishKey ks, const u8 *key, const u8 *tweak)
{
	u64 kw[WORDS+1];
	u64 tw[3];
	int i, j;
	
	for(i=0; i < WORDS; ++i)
		kw[i] = ((u64*)key)[i];
	kw[WORDS] = 0x5555555555555555;
		
	for(i=0; i < 2; ++i)
		tw[i] = ((u64*)tweak)[i];	
		
	for(i=0; i < WORDS; ++i)
		kw[WORDS] ^= kw[i];
		
	tw[2] = tw[0]^tw[1];
	
	for(i=0; i < (ROUNDS/4)+1; ++i)
	{
		for(j=0; j < WORDS; ++j)
		{
			ks->subkey[i][j] = kw[(i+j) % (WORDS+1)];
			switch(j)
			{
			case (WORDS-3):
				ks->subkey[i][j] += tw[i%3];
				break;
			case (WORDS-2):
				ks->subkey[i][j] += tw[(i+1)%3];
				break;
			case (WORDS-1):
				ks->subkey[i][j] += i;
				break;
			}
		}
	}
}

void ThreefishEncrypt(const ThreefishKey ks, u8 *in, u8 *out)
{
	u64 *p = (u64*)in;
	u64 v[WORDS], e[WORDS], f[WORDS];
	int i, d;
	
	for(i=0; i < WORDS; ++i)
		v[i] = p[i];
	
	for(d=0; d < ROUNDS; ++d)
	{
		if((d%4) == 0)
		{
			for(i=0; i < WORDS; ++i)
				e[i] = v[i] + ks->subkey[d/4][i];
		}
		else
		{
			for(i=0; i < WORDS; ++i)
				e[i] = v[i];		
		}
		
		for(i=0; i < WORDS/2; ++i)
		{
			MIX(f[i*2 + 1], f[i*2 + 0], e[i*2 + 1], e[i*2 + 0], R[d%8][i]);
		}
		
		for(i=0; i < WORDS; ++i)
			v[i] = f[Pi[i]];
	}	
	
	for(i=0; i < WORDS; ++i)
		((u64*)out)[i] = v[i] + ks->subkey[ROUNDS/4][i];	
		
}

void ThreefishDecrypt(const ThreefishKey ks, u8 *in, u8 *out)
{
	u64 *p = (u64*)in;
	u64 v[WORDS], e[WORDS], f[WORDS];
	int i, d;
	
	for(i=0; i < WORDS; ++i)
		v[i] = p[i];
	
	for(d=ROUNDS; d > 0; --d)
	{
		if((d%4) == 0)
		{
			for(i=0; i < WORDS; ++i)
				f[i] = v[i] - ks->subkey[d/4][i];
		}
		else
		{
			for(i=0; i < WORDS; ++i)
				f[i] = v[i];		
		}
		
		for(i=0; i < WORDS; ++i)
			e[i] = f[Pi[i]];
		
		for(i=0; i < WORDS/2; ++i)
		{
			IMIX(v[i*2 + 1], v[i*2 + 0], e[i*2 + 1], e[i*2 + 0], R[(d-1)%8][i]);
		}		
	}	
	
	for(i=0; i < WORDS; ++i)
		((u64*)out)[i] = v[i] - ks->subkey[0][i];	
}

// 256 bit key, 128 bit iv, doesn't matter much
void ThreefishCTR(const u8 *key, const u8 *iv, u8 *m, u8 *c, u32 len)
{
	u8 tweak[16] = {0};
	u8 workspace[32], cipher[32];
	ThreefishKey ks;
	u32 i;
	
	if(!len)
		return;
	
	ThreefishKS(ks, key, tweak);
	
	memset(workspace, 0, 32);
	memcpy(workspace + 16, iv, 16);
	
	for(;;)
	{
		ThreefishEncrypt(ks, workspace, cipher);
		
		++(((u64*)workspace)[0]);
		
		if(len <= 32)
		{
			for(i=0; i < len; ++i)
				c[i] = m[i] ^ cipher[i];
			return;
		}	
		
		for(i=0; i < 32; ++i)
			c[i] = m[i] ^ cipher[i];
			
		c += 32;
		m += 32;
		len -= 32;
	}		
}

#ifdef _TEST_

#ifdef _MSC_VER
#include <intrin.h>
#define rdtsc __rdtsc
#else
u64 rdtsc(void)
{
  u64 result;
  asm volatile(".byte 15;.byte 49;shlq $32,%%rdx;orq %%rdx,%%rax"
    : "=a" (result) ::  "%rdx");
  return result;
}
#endif

int main(int argc, char **argv)
{
	u8 key[32] = {0};
	u8 block[32] = {0};
	u8 tweak[16] = {0};
	ThreefishKey ks;
	int i;
	u64 t1, t0;
	u8 *stream=NULL;
	
	ThreefishKS(ks, key, tweak);
	
	ThreefishEncrypt(ks, block, block);
	
	for(i=0; i < 32; ++i)
		printf("%02X ", block[i]);
	printf("\n");
	
	ThreefishDecrypt(ks, block, block);
	
	for(i=0; i < 32; ++i)
		printf("%02X ", block[i]);
	printf("\n");
	
	stream = malloc(1048576); // 1 MB
	//memset(stream, 0, 1048576);
	t0 = rdtsc();
	ThreefishCTR(key, tweak, stream, stream, 1048576);
	t1 = rdtsc();
	
	printf("%f cpb for long messages.\n", (t1-t0)/1048576.0);
	
	return 0;
}


#endif

